import logging
import time

import httpx
from groq import APIConnectionError

from .config.config import load_api_key, load_model
from .log.logging_config import setup_logging

GROQ_MODEL = load_model()
GROQ_API_KEY = load_api_key()
GROQ_API_URL = "https://api.groq.com/openai/v1/chat/completions"


logger = logging.getLogger(__name__)
setup_logging()


def groq_api_call(message: str, client: httpx.Client) -> str | None:
    headers = {
        "Authorization": f"Bearer {GROQ_API_KEY}",
        "Content-Type": "application/json",
    }
    data = {
        "model": GROQ_MODEL,
        "messages": [{"role": "user", "content": message}],
        "max_tokens": 400,
        "temperature": 0.7,
    }

    max_retries, retry_delay = 3, 1
    for attempt in range(max_retries):
        try:
            response = client.post(
                GROQ_API_URL, headers=headers, json=data, timeout=15.0
            )
            response.raise_for_status()
            return response.json()["choices"][0]["message"]["content"]
        except (httpx.ConnectError, httpx.TimeoutException, APIConnectionError) as e:
            if attempt == max_retries - 1:
                logger.error(f"Connection error occured: {e}", exc_info=True)
            # Exponential backoff
            time.sleep(retry_delay * (2**attempt))
        except httpx.HTTPStatusError as e:
            if e.response.status_code >= 500:
                if attempt == max_retries - 1:
                    raise
                time.sleep(retry_delay * (2**attempt))  # Exponential backoff
            else:
                raise  # Client errors should not be retried
    return None



import argparse
import asyncio
import logging
import os
import re
import sys

import httpx
import redis
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout
from rich import print

from .chat import start_conversation_session
from .config.config import configure_api_key, configure_model, load_config, reset_config
from .config.directory import history_file
from .format import normal_bottom_toolbar, print_formatted_response
from .log.logging_config import setup_logging
from .responses import (
    generate_nofluff_response,
    generate_programming_response,
    generate_response,
)

if os.environ.get("PYTHON_DEBUG", 0) == "1":
    import debugpy

    debugpy.listen(("127.0.0.1", 5678))
    print("debugpy listening on port 5678")
    debugpy.wait_for_client()

logger = logging.getLogger(__name__)
VERSION = "0.20.4"


class InvalidInputError(Exception):
    pass


class InputTooLongError(InvalidInputError):
    pass


def get_input() -> tuple[str, argparse.Namespace]:
    parser = argparse.ArgumentParser(
        prog="LiteLookup",
        description=f"""LiteLookup: Your lightweight command-line learning companion. 
        Get simplified explanations about any concept, from general knowledge to 
        programming specifics. Use -p for programming-focused information,
        and -i for an interactive shell. Perfect for quick lookups and continuous learning sessions.
        
        version — {VERSION} """,
    )
    group = parser.add_mutually_exclusive_group()
    parser.add_argument("content", nargs="*")
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="Enters a shell session for faster lookups",
    )
    group.add_argument(
        "-d",
        "--direct",
        action="store_true",
        help="returns a no-fluff response on a programming query",
    )
    group.add_argument(
        "-c",
        "--chat",
        action="store_true",
        help="enters a conversational mode to brainstorm ideas",
    )
    parser.add_argument(
        "--version", action="version", version=f"{parser.prog} {VERSION}"
    )

    group.add_argument(
        "-p",
        "--programming",
        action="store_true",
        help="optimize response for programming concepts",
    )
    parser.add_argument(
        "--reset",
        action="store_true",
        help="Reset the API key and model configuration",
    )
    args = parser.parse_args()

    if args.reset:
        reset_config()
        return "", args

    if args.interactive and not args.content:
        return "", args

    text = " ".join(args.content)
    return (
        validate_input(text, args.interactive),
        args,
    )


def validate_input(input: str, interactive: bool) -> str:
    if not interactive and not input:
        raise InvalidInputError(
            "Input cannot be empty. Please provide a concept to check"
        )

    if len(input) > 150:
        raise InputTooLongError(
            "Text cannot be more than 150 characters long. Consider shortening."
        )

    # Input cannot contain two or more hyphens together
    if re.search(r"[-]{2,}", input):
        raise UnsupportedCharactersError(
            "Text cannot contain two or more hyphens together."
        )

    # preserve words that are ALL CAPS
    if input.isupper() or input.islower():
        return input.strip()

    return input.lower().strip()


def interactive_session(
    session_interactive: bool,
    chat: bool = False,
    direct: bool = False,
    programming: bool = False,
):
    if chat is True:
        return start_conversation_session()
    else:
        return start_normal_session(
            session_interactive=session_interactive,
            direct=direct,
            programming=programming,
        )


async def start_normal_session(
    session_interactive: bool,
    programming: bool = False,
    direct: bool = False,
):
    session = PromptSession(mouse_support=True, history=FileHistory(str(history_file)))
    with httpx.Client(
        http2=True, limits=httpx.Limits(max_keepalive_connections=5, max_connections=10)
    ) as client:
        # Set up Redis connection
        redis_client = redis.Redis(host="localhost", port=6379, db=0)

        session_timeout = 3600  # 1 hour in seconds

        while session_interactive:
            try:
                with patch_stdout():
                    user_input = await asyncio.wait_for(
                        session.prompt_async(
                            ">> lookup: ",
                            bottom_toolbar=normal_bottom_toolbar,
                        ),
                        timeout=session_timeout,
                    )
            except KeyboardInterrupt:
                continue
            except EOFError:
                break
            except asyncio.TimeoutError:
                logger.info("Session timed out due to inactivity")
                break

            text = validate_input(user_input, session_interactive)

            if text.lower() in ("q", "quit", "exit"):
                session_interactive = False
                logger.info("Exiting LiteLookup. Goodbye!")
                break

            if text and direct is True:
                response = generate_nofluff_response(text, client, redis_client)
                print_formatted_response(response)
            elif text and programming is True:
                response = generate_programming_response(text, client, redis_client)
                print_formatted_response(response)
            else:
                response = generate_response(text, client, redis_client)
                print_formatted_response(response)


def main():
    setup_logging()
    user_input, args = get_input()
    if args.reset:
        reset_config()
        return
    api_key, model = load_config()
    if api_key is None or api_key == "":
        api_key = configure_api_key()
        if api_key is None:
            print("API key is required to use this tool. Exiting.")
            sys.exit(1)
        else:
            model = configure_model()
            if model:
                print("API key and model configured. Please run your command again.")
                sys.exit(0)
    try:
        client = httpx.Client(
            http2=True,
            limits=httpx.Limits(max_keepalive_connections=5, max_connections=10),
        )
        user_input, args = get_input()
        redis_client = redis.Redis()
        if args.interactive:
            if args.programming:
                logger.info(
                    f"Switching to interactive programming mode using {model}...\n"
                )
                interactive_session(
                    args.interactive,
                    programming=True,
                )
            elif args.chat:
                logger.info(f"Switching to conversational mode using {model}...\n\n")
                asyncio.run(interactive_session(args.interactive, chat=True))
            elif args.direct:
                logger.info(
                    f"Switching to interactive no-frills mode using {model}...\n"
                )
                asyncio.run(
                    interactive_session(
                        args.interactive,
                        direct=True,
                    )
                )
            else:
                logger.info(f"Switching to interactive mode using {model}...\n")
                interactive_session(
                    args.interactive,
                    programming=False,
                    direct=False,
                )
        elif args.programming:
            logger.info(f"programming mode using {model} ...\n\n")
            response = generate_programming_response(user_input, client, redis_client)
            print_formatted_response(response)
        elif args.direct:
            logger.info(f"direct mode using {model}...\n\n")
            response = generate_nofluff_response(user_input, client, redis_client)
            print_formatted_response(response)
        else:
            logger.info(f"normal mode using {model}...\n\n")
            response = generate_response(user_input, client, redis_client)
            print_formatted_response(response)
    except (InvalidInputError, InputTooLongError) as e:
        logger.error(f"Invalid input: {str(e)} ")


if __name__ == "__main__":
    main()



import logging

import httpx
import redis

from .llm import groq_api_call
from .log.logging_config import setup_logging

logger = logging.getLogger(__name__)
setup_logging()


def generate_response(
    concept: str, client: httpx.Client, redis_client: redis.Redis
) -> str:
    try:
        cached_response = redis_client.get(concept)
        if cached_response:
            logger.info("Got a cached response\n")
            return cached_response.decode("utf-8")
        user_message = f"""Provide a concise, valuable, beginner-friendly response to the query: '{concept}'
        Your response should:
        1. Directly address the main point
        2. Offer key information or steps if applicable
        3. Be immediately useful without unnecessary background

        Adapt your response style to the query type (definition, process, comparison, etc.).
        Aim for clarity and brevity. Avoid speculation.

        Begin your response immediately without preamble.
        """
        response = groq_api_call(user_message, client)
        if response is None:
            return (
                "Sorry, there was a problem connecting to the server. Please try again"
            )
        try:
            redis_client.set(concept, response, ex=3600)
            return response
        except redis.RedisError as e:
            logger.error(f"Redis error: {e}")
            return "An error occured wjile retrieving query. Please try again"
    except redis.RedisError as e:
        logger.error(f"Redis error: {e}")
        return "An error occured wjile retrieving query. Please try again"
    except Exception as e:
        logger.error(f"An unexpected error occured: {e}.")
        return "An unexpected error occured. Check your connection and retry"


def generate_programming_response(
    concept: str, client: httpx.Client, redis_client: redis.Redis
) -> str:
    try:
        cached_response = redis_client.get(concept + "_p")  # Correct
        if cached_response:
            logger.info("Got a cached response\n")
            return cached_response.decode("utf-8")
        user_message = f"""Provide a concise yet comprehensive answer to the following programming question: '{concept}'

        Guidelines:
        1. Begin with the most relevant information directly addressing the question.
        2. Offer a brief technical explanation (2-3 sentences).
        3. List key points using numbers, covering:
           a. Basic concept or implementation
           b. Common variations or options
           c. Use cases or best practices
        4. Include short, relevant code examples where applicable.
        5. Mention a best practice or common pitfall if relevant.

        Use Markdown for code blocks. Be direct and avoid unnecessary preambles, question repetition, or filler phrases. Focus solely on the specific information requested.
        """
        response = groq_api_call(user_message, client)
        if response is None:
            return (
                "Sorry, there was a problem connecting to the server. Please try again"
            )
        try:
            redis_client.set(concept + "_p", response, ex=3600)
            return response
        except redis.RedisError as e:
            logger.error(f"Failed to cache response: {e}")
    except redis.RedisError as e:
        logger.error(f"Redis error: {e}")
        return "An error occured wjile retrieving query. Please try again"
    except Exception as e:
        logger.error(f"An unexpected error occured: {e}.")
        return "An unexpected error occured. Check your connection and retry"


def generate_nofluff_response(
    concept: str, client: httpx.Client, redis_client: redis.Redis
) -> str:
    try:
        cached_response = redis_client.get(concept + "_d")
        if cached_response:
            logger.info("Got a cached response\n")
            return cached_response.decode("utf-8")
        user_message = f"""Directly answer "{concept}" in 1-2 crisp sentences. Include only essential information. Adapt to query type (definition, step, fact, etc.) without explanation."""
        response = groq_api_call(user_message, client)
        if response is None:
            return (
                "Sorry, there was a problem connecting to the server. Please try again"
            )
        try:
            redis_client.set(concept + "_d", response, ex=3600)
            return response
        except redis.RedisError as e:
            logger.error(f"Redis error: {e}")
            return "An error occured wjile retrieving query. Please try again"
    except redis.RedisError as e:
        logger.error(f"Redis error: {e}")
        return "An error occured wjile retrieving query. Please try again"
    except Exception as e:
        logger.error(f"An unexpected error occured: {e}.")
        return "An unexpected error occured. Check your connection and retry"



import atexit
import datetime as dt
import json
import logging
import logging.config
import logging.handlers
from pathlib import Path

from litelookup.config.directory import get_app_directory

LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


def override(method):
    return method


class MyJSONFormatter(logging.Formatter):
    def __init__(
        self,
        *,
        fmt_keys: dict[str, str] | None = None,
    ):
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else {}

    @override
    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord):
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created, tz=dt.timezone.utc
            ).isoformat(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: (
                msg_val
                if (msg_val := always_fields.pop(val, None)) is not None
                else getattr(record, val)
            )
            for key, val in self.fmt_keys.items()
        }
        message.update(always_fields)

        for key, val in record.__dict__.items():
            if key not in LOG_RECORD_BUILTIN_ATTRS:
                message[key] = val

        return message


def setup_logging():
    log_file_path = get_app_directory()
    log_file = log_file_path / "litelookup.log.jsonl"

    log_file_path.mkdir(parents=True, exist_ok=True)

    base_dir = Path(__file__).resolve().parent.parent
    json_config_path = base_dir / "log" / "config.json"

    with open(json_config_path, "r") as f_in:
        config = json.load(f_in)

    config["handlers"]["file_json"]["filename"] = log_file
    logging.config.dictConfig(config)
    queue_handler = logging.getHandlerByName("queue_handler")
    if queue_handler is not None:
        queue_handler.listener.start()
        atexit.register(queue_handler.listener.stop)


class ExcludeFilter(logging.Filter):
    def __init__(self, names: list = None):
        super().__init__()
        self.names = names or []

    @override
    def filter(self, record: logging.LogRecord) -> bool | logging.LogRecord:
        return not any(record.name.startswith(name) for name in self.names)



import configparser
import os
import platform
from pathlib import Path

from prompt_toolkit import prompt
from prompt_toolkit.shortcuts import radiolist_dialog
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import Validator

from .model import GroqModel


def is_valid_APIkey(api_key: str):
    return api_key.strip().startswith("gsk_")


def is_valid_model(model_input: str) -> bool:
    # Check if model is a valid GroqModel
    return any(model_input == model.value for model in GroqModel)


def list_available_models() -> list[str]:
    return [model.value for model in GroqModel]


apikey_validator = Validator.from_callable(
    is_valid_APIkey,
    error_message="API key should start with 'gsk_'.",
    move_cursor_to_end=True,
)

model_validator = Validator.from_callable(
    is_valid_model,
    error_message="Model must be one of: " + ", ".join(list_available_models()),
    move_cursor_to_end=True,
)


def get_config_dir():
    system = platform.system()
    if system == "Windows":
        return Path(os.getenv("LOCALAPPDATA", os.path.expanduser("~"))) / "litelookup"
    elif system in ["Linux", "Darwin"]:
        return Path(os.path.expanduser("~/.config/litelookup"))
    else:
        # fallback
        return Path(os.path.expanduser("~/.litelookup"))


def get_user_key():
    print("Get your free API key from https://console.groq.com/keys")
    try:
        api_key = prompt(
            "Paste API key here: ",
            validator=apikey_validator,
            validate_while_typing=False,
        ).strip()
        return api_key if api_key else None
    except (KeyboardInterrupt, EOFError):
        return None


def get_user_model() -> str | None:
    model_choices = [(model, model.value) for model in GroqModel]

    custom_style = Style.from_dict(
        {
            "dialog": "bg:#2b2b2b",
            "dialog frame.label": "bg:#333333 #ffffff bold",
            "dialog.body": "bg:#2b2b2b #ffffff",
            "dialog shadow": "bg:#1c1c1c",
            "radio": "bg:#2b2b2b #ffffff",
            "radio-selected": "bg:#005fff #ffffff",
            "button": "bg:#555555 #ffffff",
            "button.focused": "bg:#005fff #ffffff",
            "text-area": "bg:#2b2b2b #ffffff",
        }
    )
    result = radiolist_dialog(
        title="Select a Groq model",
        text="Choose one of the available models:",
        values=model_choices,
        style=custom_style,
    ).run()
    if result:
        return result.value
    else:
        print("No model selected")
        return None


def create_config_file(config_dir: Path) -> Path:
    config_dir.mkdir(parents=True, exist_ok=True)
    config_file = config_dir / "config.ini"
    return config_file


def store_api_key(api_key: str) -> bool:
    config_file = create_config_file(get_config_dir())

    try:
        with config_file.open("w") as f:
            f.write(f"[env]\nGROQ_API_KEY={api_key}\n")
        os.chmod(config_file, 0o600)
        return True
    except Exception as e:
        print(f"Error storing API key: {str(e)}")
        return False


def store_model(model: str) -> bool:
    config_file = create_config_file(get_config_dir())

    try:
        with config_file.open("a") as f:
            f.write(f"GROQ_MODEL={model}\n")
        os.chmod(config_file, 0o600)
        return True
    except Exception as e:
        print(f"error storing model: {str(e)}")
        return False


def get_config_file() -> Path:
    return get_config_dir() / "config.ini"


def load_api_key() -> str | None:
    config = configparser.ConfigParser()
    config_file = get_config_file()
    config.read(config_file)
    try:
        api_key = config["env"]["GROQ_API_KEY"]
    except KeyError:
        return None
    return api_key


def load_model() -> str | None:
    config = configparser.ConfigParser()
    config_file = get_config_file()
    config.read(config_file)
    try:
        model = config["env"]["GROQ_MODEL"]
    except KeyError:
        return None
    return model


def load_config():
    api_key = load_api_key()
    model = load_model()
    return api_key, model


def configure_api_key():
    while True:
        api_key = get_user_key()
        if api_key is None:
            print("API key configuration cancelled.")
            return None
        if store_api_key(api_key):
            return api_key
        print("Error storing API key. Please try again.")


# set the apikey entry to an empty string
def reset_config():
    config = configparser.ConfigParser()
    config_file = get_config_dir() / "config.ini"
    if config_file.exists():
        config.read(config_file)
        with config_file.open("w+") as cf:
            cf.write("[env]\nGROQ_API_KEY=")
    else:
        return


def configure_model():
    while True:
        model = get_user_model()
        if model is None:
            print("Model configuration cancelled.")
            return None
        if store_model(model):
            return model
        print("Error storing model. Please try again.")
